#ifndef AUTOPALETTE_H
#define AUTOPALETTE_H

#include <cv_header.h>
#include <iostream>
#include <math.h>

using namespace std;

//inline void lab2rgb( double l_s, double a_s, double b_s, double& R, double& G, double& B )
//{
//	double var_Y = ( l_s + 16. ) / 116.;
//	double var_X = a_s / 500. + var_Y;
//	double var_Z = var_Y - b_s / 200.;
//
//	if ( pow(var_Y,3) > 0.008856 ) var_Y = pow(var_Y,3);
//	else                      var_Y = ( var_Y - 16. / 116. ) / 7.787;
//	if ( pow(var_X,3) > 0.008856 ) var_X = pow(var_X,3);
//	else                      var_X = ( var_X - 16. / 116. ) / 7.787;
//	if ( pow(var_Z,3) > 0.008856 ) var_Z = pow(var_Z,3);
//	else                      var_Z = ( var_Z - 16. / 116. ) / 7.787;
//
//	double X = 95.047 * var_X ;    //ref_X =  95.047     Observer= 2, Illuminant= D65
//	double Y = 100.000 * var_Y  ;   //ref_Y = 100.000
//	double Z = 108.883 * var_Z ;    //ref_Z = 108.883
//
//
//	var_X = X / 100. ;       //X from 0 to  95.047      (Observer = 2, Illuminant = D65)
//	var_Y = Y / 100. ;       //Y from 0 to 100.000
//	var_Z = Z / 100. ;      //Z from 0 to 108.883
//
//	double var_R = var_X *  3.2406 + var_Y * -1.5372 + var_Z * -0.4986;
//	double var_G = var_X * -0.9689 + var_Y *  1.8758 + var_Z *  0.0415;
//	double var_B = var_X *  0.0557 + var_Y * -0.2040 + var_Z *  1.0570;
//
//	if ( var_R > 0.0031308 ) var_R = 1.055 * pow(var_R , ( 1 / 2.4 ))  - 0.055;
//	else                     var_R = 12.92 * var_R;
//	if ( var_G > 0.0031308 ) var_G = 1.055 * pow(var_G , ( 1 / 2.4 ) )  - 0.055;
//	else                     var_G = 12.92 * var_G;
//	if ( var_B > 0.0031308 ) var_B = 1.055 * pow( var_B , ( 1 / 2.4 ) ) - 0.055;
//	else                     var_B = 12.92 * var_B;
//
//	R = var_R;
//	G = var_G;
//	B = var_B;
//
//}

class CIELAB
{
public:
	CIELAB()
	{
		L = 0;
		A = 0;
		B = 0;
	}

	CIELAB(double l, double a, double b)
	{
		L = l;
		A = a;
		B = b;
	}

	double SqDist(CIELAB lab)
	{
		double ldiff = this->L - lab.L;
		double adiff = this->A - lab.A;
		double bdiff = this->B - lab.B;
		return pow(ldiff, 2.0) + pow(adiff, 2.0) + pow(bdiff, 2.0);
	}

public:
	double L;
	double A;
	double B;
};

//class RGB
//{
//public:
//	RGB()
//	{
//		R = 0;
//		G = 0;
//		B = 0;
//	}
//	RGB(double r, double g, double b)
//	{
//		R = r;
//		G = g;
//		B = b;
//	}
//public:
//	double R;
//	double G;
//	double B;
//}; 


class Cluster
{
public:
	Cluster()
	{
		norm_weight = 0;
		count = 0;
		id = -1;
	}
	~Cluster()
	{
	}
public:
	CIELAB MeanColor()
	{
		if (count == 0)
			return lab;
		//return CIELAB(sumlab.L / (double)count, sumlab.A / (double)count, sumlab.B / (double)count);
		return CIELAB(sumlab.L / norm_weight, sumlab.A / norm_weight, sumlab.B / norm_weight);
	}

	void AddColor(CIELAB color, double weight)
	{
		sumlab.L += weight * color.L;
		sumlab.A += weight * color.A;
		sumlab.B += weight * color.B;
		norm_weight += weight;
		count += 1;
	}

	void Reset()
	{
		lab = MeanColor();
		count = 0;
		norm_weight = 0;
		sumlab = CIELAB(0, 0, 0);
	}
public:
	int id;
	CIELAB lab;
	int count;
	double norm_weight;
private:
	CIELAB sumlab;
};


class Clustering
{
public:
	static void ColorQuantization(Mat img, int bins_num, vector<CIELAB> &meanLabColor, 
		vector<double> &bucketWeight)
	{
		double eps = 1e-10;

		int width = img.cols;
		int height = img.rows;

		double total = (double) (width * height);

		int total_bins_num = bins_num * bins_num * bins_num;
		vector<vector<CIELAB> > colorBucket(total_bins_num);

		Mat nor_img, lab_img;
		img.convertTo(nor_img, CV_32FC3, 1/255.0f);
		cvtColor(nor_img, lab_img, CV_BGR2Lab);

		int r_ind = 0, g_ind = 0, b_ind = 0;
		double r_Val = 0, g_Val = 0, b_Val = 0;
		double l_val = 0, a_val = 0, b_val = 0;

		for (int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				Vec3b pixelVal = img.at<Vec3b>(i, j);
				Vec3f labVal = lab_img.at<Vec3f>(i, j);

				r_ind = (int)pixelVal[2] / 16;
				g_ind = (int)pixelVal[1] / 16;
				b_ind = (int)pixelVal[0] / 16;

				int rgb_index = r_ind * bins_num * bins_num + g_ind * bins_num + b_ind;

				CIELAB lab(labVal[0], labVal[1], labVal[2]);
				colorBucket[rgb_index].push_back(lab);
			}
		}

		double sum_l = 0, sum_a = 0, sum_b = 0;
		for (int i = 0; i < total_bins_num; i++)
		{
			sum_l = 0; sum_a = 0; sum_b = 0;
			double element_count = (int)colorBucket[i].size() + eps;
			for (int j = 0; j < (int)colorBucket[i].size(); j++)
			{
				sum_l += colorBucket[i][j].L;
				sum_a += colorBucket[i][j].A;
				sum_b += colorBucket[i][j].B;
			}

			CIELAB lab(sum_l / element_count, sum_a / element_count, sum_b / element_count);
			meanLabColor.push_back(lab);
			//double weight = element_count / total;
			double weight = element_count;
			bucketWeight.push_back(weight);
		}
	}

	static vector<Cluster> InitializeColorSeeds(vector<CIELAB> meanLabColor, 
		vector<double> bucketWeight, int k)
	{
		// initialize k + 1 centers
		vector<Cluster> initialCenters;

		// initialize the black seed
		CIELAB black_seed(0, 0, 0);
		Cluster redundant_cluster;
		redundant_cluster.id = 0;
		redundant_cluster.lab = black_seed;
		initialCenters.push_back(redundant_cluster);

		// determine following k seeds
		double sigma_alpha = 80.0;
		vector<double> temp_weight(bucketWeight);
		for (int i = 0; i < k; i++)
		{
			auto maxWeight = std::max_element(temp_weight.begin(), temp_weight.end());
			int maxWeightPosition = std::distance(temp_weight.begin(), maxWeight);
			Cluster cluster_temp;
			cluster_temp.id = i + 1;
			cluster_temp.lab = meanLabColor[maxWeightPosition];
			initialCenters.push_back(cluster_temp);

			// exclude the current max weight in the coming round by setting its weight to zero
			temp_weight[maxWeightPosition] = 0;

			// update all other weights in bucketWeight
			for (int j = 0; j < (int)temp_weight.size(); j++)
			{
				if (j == maxWeightPosition)
					continue;
				else
				{
					double l_diff = meanLabColor[j].L - meanLabColor[maxWeightPosition].L;
					double a_diff = meanLabColor[j].A - meanLabColor[maxWeightPosition].A;
					double b_diff = meanLabColor[j].B - meanLabColor[maxWeightPosition].B;
					double lab_dist = l_diff * l_diff + a_diff * a_diff + b_diff * b_diff;
					double factor = 1 - exp(-lab_dist / (sigma_alpha * sigma_alpha));
					temp_weight[j] *= factor;
				}
			}
		}

		return initialCenters;
	}

	static void KMeansColorThemeExtraction(vector<CIELAB> colors, vector<Cluster> &seeds, 
		vector<double> weight, int maxIters=50)
	{
		// cluster colors given seeds
		// go through colors
		int* assignments = new int[(int)colors.size()];
		int changes = 1;

		// store count of each cluster
		//int *t_count = new int [(int)seeds.size()];

		for (int t = 0; t < maxIters; t++)
		{
			changes = 0;

			for (int i = 0; i < (int)colors.size(); i++)
			{
				double bestDist = numeric_limits<double>::max();
				int bestSeed = -1;

				// go through seeds and pick best one
				for(int j = 0; j < (int)seeds.size(); j++)
				{
					double dist = seeds[j].lab.SqDist(colors[i]);
					if (dist < bestDist)
					{
						bestDist = dist;
						bestSeed = seeds[j].id;
						//bestSeed = j;
					}
				}

				// check the assignment
				if (assignments[i] != bestSeed)
					changes++;
				assignments[i] = bestSeed;
				seeds[bestSeed].AddColor(colors[i], weight[i]);
			}

			// update means
			for (int i = 0; i < (int)seeds.size(); i++)
			{
				//// if seed is starved, try again, just pick a random color
				//if (seeds[i].count == 0)
				//{
				//	srand((int)time(0));
				//	seeds[i].lab = colors[random(colors.size()-1)];
				//	changes++;
				//}
				//// return max of double;
				//else
				//{
				//	t_count[i] = seeds[i].count;
				//	seeds[i].Reset();
				//}
				//t_count[i] = seeds[i].count;
				seeds[i].Reset();
			}

			if (changes == 0)
				break;
		}

		//for (int h = 0; h < (int)seeds.size(); h++)
		//	seeds[h].count = t_count[h];

		//delete [] t_count;
		delete [] assignments;
	}
};

//bool comp(const Cluster &c1, const Cluster &c2)
//{
//	return c1.count > c2.count;
//}

class AutoPaletteExtraction{
public:
	AutoPaletteExtraction(const char *path, int k)
	{
		srcImg = imread(path,CV_LOAD_IMAGE_COLOR);

		width = srcImg.cols;
		height = srcImg.rows;

		kNum = k;

		//labPalette = new CIELAB[kNum];
	}
	~AutoPaletteExtraction()
	{
		//delete [] labPalette;
		//delete [] rgbPalette;

		//labPalette = NULL;
		//rgbPalette = NULL;
	}
public:
	void extractPalette()
	{
		int bins_num = 16;

		/* color quantization */
		vector<CIELAB> meanColor;
		vector<double> weight;

		Clustering::ColorQuantization(srcImg, bins_num, meanColor, weight);
		cout<<"color quantization finished ..."<<endl;

		/* color seeds initialization */
		vector<Cluster> vec_seed = Clustering::InitializeColorSeeds(meanColor, weight, kNum);
		cout<<"color seeds initialized ..."<<endl;

		/* k-means clustering */
		Clustering::KMeansColorThemeExtraction(meanColor, vec_seed, weight);
		cout<<"k-means clustering finished ..."<<endl;

		//renderPalette(vec_seed);

		// convert from lab to bgr
		Mat labPalette(1, kNum, CV_32FC3);
		for (int i = 0; i < kNum; i++)
		{
			labPalette.at<Vec3f>(0, i)[0] = (float)vec_seed[i].lab.L;
			labPalette.at<Vec3f>(0, i)[1] = (float)vec_seed[i].lab.A;
			labPalette.at<Vec3f>(0, i)[2] = (float)vec_seed[i].lab.B;
		}

		cvtColor(labPalette, bgrPalette, CV_Lab2BGR);

		//bgrPalette.at<Vec3f>(0, 0) = Vec3f(106 / 255.0f, 148 / 255.0f, 178 / 255.0f);
		//bgrPalette.at<Vec3f>(0, 1) = Vec3f(130 / 255.0f, 154 / 255.0f, 156 / 255.0f);
		//bgrPalette.at<Vec3f>(0, 2) = Vec3f(49 / 255.0f, 154 / 255.0f, 197 / 255.0f);
		//bgrPalette.at<Vec3f>(0, 3) = Vec3f(101 / 255.0f, 161 / 255.0f, 169 / 255.0f);
		//bgrPalette.at<Vec3f>(0, 4) = Vec3f(101 / 255.0f, 161 / 255.0f, 169 / 255.0f);

	}

	//void renderPalette(vector<Cluster> palette)
	//{
	//	int res_rows = 100;
	//	int res_cols = 100 * kNum;

	//	//sort(palette.begin(),palette.end(),comp);  

	//	/* store sorted CIELAB themes */
	//	for (int i = 0; i < (int) kNum; i++)
	//	{
	//		labPalette[i] = palette[i+1].lab;
	//	}

	//	Mat render_res(res_rows, res_cols, CV_32FC3);
	//	for (int i = 0; i < kNum; i++)
	//	{
	//		for (int j = 0; j < 100; j++)
	//		{
	//			for (int k = 0; k < 100; k++)
	//			{
	//				render_res.at<Vec3f>(j, i*100 + k)[0] = (float)labPalette[i].L;
	//				render_res.at<Vec3f>(j, i*100 + k)[1] = (float)labPalette[i].A;
	//				render_res.at<Vec3f>(j, i*100 + k)[2] = (float)labPalette[i].B;
	//			}
	//		}
	//		//cout<<theme[i].count<<endl;
	//	}

	//	//Mat rgbRes;
	//	cvtColor(render_res, resPalette, CV_Lab2BGR);
	//	resPalette.convertTo(resPalette, CV_32FC3, 255.0);

	//	//imwrite("E:/colorPalette.png", rgbRes);
	//	//return rgbRes;
	//}

	Mat getBGRPalette()
	{
		return bgrPalette;
	}

private:
	Mat srcImg;
	int width;
	int height;
	int kNum;

	//CIELAB *labPalette;
	Mat bgrPalette;
	//RGB* rgbPalette;
};

#endif  // AutoPalette.h