#ifndef KCOLORTHEME_H
#define KCOLORTHEME_H

#include <cv_header.h>
#include <iostream>
#include <math.h>
#include <vector>
#include <algorithm>
#include <limits>
#include <time.h>
#include "PBCG.h"
#include "LLE.h"
#include "calcAffinity.h"
#include "AutoPalette.h"

using namespace std;

class ThemeExtraction
{
public:
	ThemeExtraction(const char *path, int k)
	{
		srcImg = imread(path, CV_LOAD_IMAGE_COLOR);
		//Mat rgbImg(srcImg.rows, srcImg.cols,CV_32FC3);
		srcImg.convertTo(rgbImg, CV_32FC3, 1/255.0);
		cvtColor(rgbImg, labImg, CV_BGR2Lab);

		width = srcImg.cols;
		height = srcImg.rows;

		k_num = k;
		//labThemes = new CIELAB [k_num];
		//rgbThemes = new RGB [k_num];

		/* allocate memory for weight array */
		weight = new double [k_num * width * height];
		dissimilar_weight = new double [k_num * width * height];
		memset(weight, 0, sizeof(double) * k_num * height * width);
		memset(dissimilar_weight, 0, sizeof(double) * k_num * height * width);

		// palette extraction
		AutoPaletteExtraction ape(path, k_num);
		ape.extractPalette();
		rgbThemes = ape.getBGRPalette();
		cvtColor(rgbThemes, labdst, CV_BGR2Lab);

		
		split(labImg, channel);
		/* filtering the luminance channel */
		//blur(channel[0], channel[0], Size(7, 7));
		//GaussianBlur(channel[0], channel[0], Size(7, 7), 0, 0);
	}
	~ThemeExtraction()
	{
		//delete [] labThemes;
		//delete [] rgbThemes;

		//labThemes = NULL;
		//rgbThemes = NULL;

		delete [] weight;
		weight = NULL;

		delete [] dissimilar_weight;
		dissimilar_weight = NULL;
	}
public:
	Mat getRGBThemes() { return rgbThemes; }

	double calcManifoldDiff(double *w, Manifold **pt, int nnCount, int pageID, int x_coord, int y_coord)
	{
		int pageCount = pageID * width * height;
		int ind = pageCount + y_coord * width + x_coord;
		double manifold_diff = w[ind];
		int local_ind = y_coord * width + x_coord;
		for (int i = 0; i < nnCount; i++)
		{
			int local_w = pt[local_ind][i].indx % width;
			int local_h = pt[local_ind][i].indx / width;
			int global_ind = pageCount + local_h * width + local_w;

			manifold_diff = manifold_diff - w[global_ind] * pt[local_ind][i].weight;
		}

		return fabs(manifold_diff);
	}

	void computeDissimilarWeight()
	{
		double chroma_sigma = 80.0;

		double cur_sum = 0;

		for (int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				cur_sum = 0;
				for (int k = 0; k < k_num; k++)
				{
					int index = k * width * height + i * width + j;

					double l_diff = labImg.at<Vec3f>(i, j)[0] - labdst.at<Vec3f>(0, k)[0];
					double a_diff = labImg.at<Vec3f>(i, j)[1] - labdst.at<Vec3f>(0, k)[1];
					double b_diff = labImg.at<Vec3f>(i, j)[2] - labdst.at<Vec3f>(0, k)[2];
					double chroma_dist = a_diff * a_diff + b_diff * b_diff;
					double total_dist = l_diff * l_diff + a_diff * a_diff + b_diff * b_diff;
				
					dissimilar_weight[index] = exp(-total_dist / (chroma_sigma * chroma_sigma));
					cur_sum += dissimilar_weight[index];
				}

				// normalize the dissimilarity weight to [0, 1]
				for (int k = 0; k < k_num; k++)
				{
					int index = k * width * height + i * width + j;
					dissimilar_weight[index] /= cur_sum;
					dissimilar_weight[index] = 1 - dissimilar_weight[index];
				}
			}
		}
	}


	void colorThemeDecompose()  /* IRLS solver */
	{
		/* locally linear embedding manifold construction */
		int nnCount = 50;     
		LLE le(srcImg, nnCount);
		le.initialization();
		cout<<"lle initialization finished !"<<endl;
		le.KNNSearch();
		cout<<"KNN search finished !"<<endl;
		le.calcLLEWeights();
		cout<<"manifold weights computed !"<<endl;
		le.reArranging();	
		cout<<"manifold rearranging finished !"<<endl;
		Manifold** manifold_pointer = le.getManifoldPointer();

		// compute the dissimilarity weight
		computeDissimilarWeight();
		cout<<"Dissimilarity weight computed !"<<endl;

		//////////////////////////////////////
		double eps = 1e-6;

		double lamda_nn = 1e3;     /* non-negative term coefficient*/ 
		double lamda_mm = 1e-2;    /* smoothness term coefficient */
		double lamda_sp = 1e-2;     /* L1 regularization term coefficient */

		/* init all the weight terms */
		int pageCount = width * height;
		double * pusai_sparse = new double [k_num * pageCount];
		double * pusai_non_negative = new double [k_num * pageCount];
		double * pusai_manifold = new double [k_num * pageCount];
		for (int i = 0; i < k_num * pageCount; i++)
		{
			//pusai_sparse[i] = 1.0;
			pusai_sparse[i] = dissimilar_weight[i];
			pusai_non_negative[i] = 1.0;
			pusai_manifold[i] = 1.0;
		}

		int iterations = 10;
		const int total = k_num * pageCount;
		int unknown_num = k_num + nnCount;

		double *sa = new double [total * unknown_num];
		unsigned long *ija = new unsigned long [total * unknown_num];
		double *b = new double [total];
		double *x = new double [total];

		memset(sa, 0, sizeof(double) * total * unknown_num);
		memset(ija, 0, sizeof(unsigned long) * total * unknown_num);
		memset(b, 0, sizeof(double) * total);
		memset(x, 0, sizeof(double) * total);

		cout<<"total unknowns : "<<total<<" ..."<<endl;
		cout<<"iterations : "<<endl;

		for (int iter = 0; iter < iterations; iter++)
		{
			ija[0] = total + 1;
			int k = total;
			int index = 0;

			for (int i = 0; i < k_num; i++)
			{
				for (int j = 0; j < height; j++)
				{
					for (int h = 0; h < width; h++)
					{
						index = i * pageCount + j * width + h;  
						Vec3f pixelVal = labImg.at<Vec3f>(j, h);

						sa[index] = labdst.at<Vec3f>(0, i)[2] * labdst.at<Vec3f>(0, i)[2] + 
							labdst.at<Vec3f>(0, i)[1] * labdst.at<Vec3f>(0, i)[1] + pusai_manifold[index] * 
							lamda_mm + lamda_sp * pusai_sparse[index] + lamda_nn * pusai_non_negative[index];

						b[index] = pixelVal[2] * labdst.at<Vec3f>(0, i)[2] + 
							pixelVal[1] * labdst.at<Vec3f>(0, i)[1];

						/* unknowns before page i */
						for(int ii = 0; ii < i; ii++)
						{
							k++;
							sa[k] = labdst.at<Vec3f>(0, i)[2] * labdst.at<Vec3f>(0, ii)[2] + 
								labdst.at<Vec3f>(0, i)[1] * labdst.at<Vec3f>(0, ii)[1];
							ija[k] = index - (i - ii) * pageCount;
						}

						/* unknowns within page i */
						int local_index = j * width + h;
						for(int ii = 0; ii < nnCount; ii++)
						{
							k++;
							sa[k] = -lamda_mm * pusai_manifold[index] * manifold_pointer[local_index][ii].weight;
							ija[k] = i * pageCount + manifold_pointer[local_index][ii].indx;
						}

						/* unknowns behind page i */
						for(int ii = i + 1; ii < k_num; ii++)
						{
							k++;
							sa[k] = labdst.at<Vec3f>(0, i)[2] * labdst.at<Vec3f>(0, ii)[2] + 
								labdst.at<Vec3f>(0, i)[1] * labdst.at<Vec3f>(0, ii)[1];
							ija[k] = index + (ii - i) * pageCount;
						}

						ija[index + 1] = k + 1;
					}
				}
			}

			//cout<<"sa and ija constructed ..."<<endl;

			// exploit pbcg to solve the linear system
			int mpcgx_iter;
			double mpcgx_error;
			PBCG mpcgx(sa,ija);

			mpcgx.linbcg(x,b,total,mpcgx_iter,mpcgx_error,3,1e-5,1000);

			cout<<iter+1<<"-th iteration finished ";

			/* update weight */
			for (int i = 0; i < k_num; i++)
			{
				for (int j = 0; j < height; j++)
				{
					for (int h = 0; h < width; h++)
					{
						int ind = i * pageCount + j * width + h;
						//double gradient = calcGradient(x, i, h, j);
						double manifold_diff = calcManifoldDiff(x, manifold_pointer, nnCount, i, h, j);

						pusai_sparse[ind] = dissimilar_weight[ind] / (fabs(x[ind]) + eps);
						pusai_manifold[ind] = 1.0 / (manifold_diff + eps);
						//pusai_manifold[ind] = 1.0;
						
						if (x[ind] >= 0)
							pusai_non_negative[ind] = 0;
						else 
							pusai_non_negative[ind] = 1.0 / (fabs(x[ind]) + eps);  
					}
				}
			}

			cout<<"and weights updated ... "<<endl;
		}

		/* store the result */ 
		memcpy(weight, x, sizeof(double) * total);

		/* free memory */
		delete [] pusai_sparse;
		delete [] pusai_non_negative;
		delete [] pusai_manifold;

		delete [] sa;
		delete [] ija;
		delete [] x;
		delete [] b;
	}

	void testDecomposition()
	{
		Mat lab_res(height, width, CV_32FC3);

		for (int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				double r = 0, g = 0, b = 0;
				double ll = 0, aa = 0, bb = 0;

				for (int h = 0; h < k_num; h++)
				{
					int ind = h * width * height + i * width + j;

					ll = labImg.at<Vec3f>(i, j)[0];
					aa += labdst.at<Vec3f>(0, h)[1] * weight[ind];
					bb += labdst.at<Vec3f>(0, h)[2] * weight[ind];

					//lab2rgb(ll, aa, bb, r, g, b);
				}

				lab_res.at<Vec3f>(i, j)[0] = (float)(ll);
				lab_res.at<Vec3f>(i, j)[1] = (float)(aa);
				lab_res.at<Vec3f>(i, j)[2] = (float)(bb);
			}
		}

		Mat bgr_res;
		cvtColor(lab_res, bgr_res, CV_Lab2BGR);
		bgr_res.convertTo(bgr_res, CV_32FC3, 255.0f);
		imwrite("E:/reconstructed.png", bgr_res);
	}

	Mat colorThemeTransfer(Mat input_theme)
	{
		Mat lab_res(height, width, CV_32FC3);

		vector<Mat> weight_map;
		for (int i = 0; i < k_num; i++)
		{
			Mat temp(height, width, CV_32F);
			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int ind = i * width * height + j * width + k;
					temp.at<float>(j, k) = weight[ind];
				}
			}
			//blur(temp, temp, Size(15, 15));
			//GaussianBlur(temp, temp, Size(7, 7), 0);
			weight_map.push_back(temp);
		}

		for (int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				double r = 0, g = 0, b = 0;
				double ll = 0, aa = 0, bb = 0;

				//ll = channel[0].at<float>(i, j);
				//aa = input_theme.at<Vec3f>(0, 0)[1] * weight_map[0].at<float>(i, j);
				//bb = input_theme.at<Vec3f>(0, 0)[2] * weight_map[0].at<float>(i, j);

				for (int h = 0; h < k_num; h++)
				{
					int ind = h * width * height + i * width + j;

					//ll = labImg.at<Vec3f>(i, j)[0];
					ll = channel[0].at<float>(i, j);
					//aa += input_theme.at<Vec3f>(0, h)[1] * weight[ind];
					//bb += input_theme.at<Vec3f>(0, h)[2] * weight[ind];

					aa += input_theme.at<Vec3f>(0, h)[1] * weight_map[h].at<float>(i, j);
					bb += input_theme.at<Vec3f>(0, h)[2] * weight_map[h].at<float>(i, j);


					//lab2rgb(ll, aa, bb, r, g, b);
				}

				/*res.at<Vec3f>(i, j)[0] = (float)(min(max(b, .0), 1.0));
				res.at<Vec3f>(i, j)[1] = (float)(min(max(g, .0), 1.0));
				res.at<Vec3f>(i, j)[2] = (float)(min(max(r, .0), 1.0));*/

				lab_res.at<Vec3f>(i, j)[0] = (float)(ll);
				lab_res.at<Vec3f>(i, j)[1] = (float)(aa);
				lab_res.at<Vec3f>(i, j)[2] = (float)(bb);
			}
		}

		Mat bgr_res;
		cvtColor(lab_res, bgr_res, CV_Lab2BGR);
		//bgr_res.convertTo(bgr_res, CV_32FC3, 255.0f);

		return bgr_res;
	}

	Mat computeGradient(Mat src)
	{
		Mat grad = Mat::zeros(src.rows, src.cols, CV_32F);

		for (int i = 0; i < src.rows - 1; i++)
		{
			for (int j = 0; j < src.cols - 1; j++)
			{
				float dx = src.at<float>(i, j) - src.at<float>(i, j+1);
				float dy = src.at<float>(i, j) - src.at<float>(i+1, j);
				float grad_val = sqrt(dx * dx + dy * dy);
				grad.at<float>(i, j) = grad_val;
			}
		}

		return grad;
	}

	void Decolorization()
	{
		double eps = 1e-8;
		Mat gray_res(height, width, CV_32F);

		for(int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				float r = rgbImg.at<Vec3f>(i, j)[2];
				float g = rgbImg.at<Vec3f>(i, j)[1];
				float b = rgbImg.at<Vec3f>(i, j)[0];
				float intensity = (20.0f * r + 40.0f * g + b) / 61.0f;
				gray_res.at<float>(i, j) = intensity;
			}
		}

		Mat nor_gray;
		gray_res.convertTo(nor_gray, CV_32F, 255.0f);
		imwrite("E:/naive_lum.png", nor_gray);

		vector<Mat> grad_map;
		vector<Mat> weight_map;
		weight_map.push_back(gray_res);

		for (int i = 0; i < k_num; i++)
		{
			int sIndex = i * width * height;
			double max_weight = *std::max_element(weight + sIndex, weight + sIndex + width * height);
			double min_weight = *std::min_element(weight + sIndex, weight + sIndex + width * height);

			//double r = 0, g = 0, b = 0;
			double ll = 0, aa = 0, bb = 0;

			Mat temp = Mat::zeros(height, width, CV_32F);

			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int dIndex = sIndex + j * width + k;

					ll = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);
					temp.at<float>(j, k) = ll;
				}
			}

			//bilateralFilter(temp, temp, 7, )
			//Mat uchar_temp;
			//temp.convertTo(uchar_temp, CV_8U, 255);
			//medianBlur(uchar_temp, uchar_temp, 7);
			//uchar_temp.convertTo(temp, CV_32F, 1.0f / 255.0);

			weight_map.push_back(temp);
			//Mat grad = computeGradient(temp);
			//grad_map.push_back(grad);

			//Mat bgrMap;
			//cvtColor(weightMap, bgrMap, CV_Lab2BGR);
			//bgrMap.convertTo(bgrMap, CV_32FC3, 255.0f);
			//imwrite(buf, bgrMap);
		}

		//Mat grad_src = computeGradient(gray_res);
		//grad_map.push_back(grad_src);
		//weight_map.push_back(gray_res);

		Mat gray_add = Mat::zeros(height, width, CV_32F);

		for (int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				//float max_grad = -1;
				//float gray_val = 0;
				for (int k = 0; k < 4; k++)
				{
					/*if (grad_map[k].at<float>(i, j) >= max_grad)
					{
						max_grad = grad_map[k].at<float>(i, j);
						gray_val = weight_map[k].at<float>(i, j);
					}*/
					gray_add.at<float>(i, j) += weight_map[k].at<float>(i, j);
				}
			}
		}


		/*for (int i = 0; i < height; i++)
		{
			for (int j = 0; j < width; j++)
			{
				float cur_val = gray_add.at<float>(i, j);
				gray_add.at<float>(i, j) = (cur_val - min_val) / (max_val - min_val + eps);
			}
		}*/

		gray_add.convertTo(gray_add, CV_32F, 255.0f);
		imwrite("E:/decolorization.png", gray_add);
	}

	void visualizeWeightMaps(char *path)
	{
		char buf[256];
		double eps = 1e-8;

		Mat weightMap(height, width, CV_32FC3);

		for (int i = 0; i < k_num; i++)
		{
			sprintf_s(buf, 256, "%s%d.png", path, i+1);

			int sIndex = i * width * height;
			double max_weight = *std::max_element(weight + sIndex, weight + sIndex + width * height);
			double min_weight = *std::min_element(weight + sIndex, weight + sIndex + width * height);

			double r = 0, g = 0, b = 0;
			double ll = 0, aa = 0, bb = 0;

			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int dIndex = sIndex + j * width + k;

					ll = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);
					//ll = labdst.at<Vec3f>(0, i)[0];
					aa = labdst.at<Vec3f>(0, i)[1];
					bb = labdst.at<Vec3f>(0 ,i)[2];

					weightMap.at<Vec3f>(j, k)[0] = (float)(ll);
					weightMap.at<Vec3f>(j, k)[1] = (float)(aa);
					weightMap.at<Vec3f>(j ,k)[2] = (float)(bb);
				}
			}

			Mat bgrMap;
			cvtColor(weightMap, bgrMap, CV_Lab2BGR);
			bgrMap.convertTo(bgrMap, CV_32FC3, 255.0f);
			imwrite(buf, bgrMap);
		}
	}

	Mat ColorMapJet(Mat src)
	{
		assert(src.channels() == 1);

		Mat dst(src.rows, src.cols, CV_8UC3);


		for (int i = 0;i < src.rows; i++)
		{
			for (int j = 0; j < src.cols; j++)
			{
				float norVal = src.at<float>(i, j);
				double db = 0, dg = 0, dr = 0;

				if (norVal < 0.1242) { 
					db = 0.504 + ((1.-0.504) / 0.1242)*norVal; 
					dg = dr = 0.; 
				} else if (norVal < 0.3747) { 
					db = 1.; 
					dr = 0.; 
					dg = (norVal - 0.1242) * (1. / (0.3747-0.1242)); 
				} else if (norVal < 0.6253) { 
					db = (0.6253 - norVal) * (1. / (0.6253-0.3747)); 
					dg = 1.; 
					dr = (norVal - 0.3747) * (1. / (0.6253-0.3747)); 
				} else if (norVal < 0.8758) { 
					db = 0.; 
					dr = 1.; 
					dg = (0.8758 - norVal) * (1. / (0.8758-0.6253)); 
				} else { 
					db = 0.; 
					dg = 0.; 
					dr = 1. - (norVal - 0.8758) * ((1.-0.504) / (1.-0.8758)); 
				}

				dst.at<Vec3b>(i, j)[0] = (uchar)(255.0 * db);
				dst.at<Vec3b>(i, j)[1] = (uchar)(255.0 * dg);
				dst.at<Vec3b>(i, j)[2] = (uchar)(255.0 * dr);
			}
		}

		return dst;
	}

	void visWeightMaps(char *path)
	{
		char buf[256];
		double eps = 1e-8;

		//Mat weightMap(height, width, CV_32FC3);
		Mat grayWeight(height, width, CV_32FC1);

		for (int i = 0; i < k_num; i++)
		{
			sprintf_s(buf, 256, "%s%d.png", path, i+1);

			int sIndex = i * width * height;
			double max_weight = *std::max_element(weight + sIndex, weight + sIndex + width * height);
			double min_weight = *std::min_element(weight + sIndex, weight + sIndex + width * height);

			//double r = 0, g = 0, b = 0, src_intensity = 0, dst_intensity = 0;

			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int dIndex = sIndex + j * width + k;
					grayWeight.at<float>(j, k) = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);

					//b = rgbThemes.at<Vec3f>(0, i)[0];
					//g = rgbThemes.at<Vec3f>(0, i)[1];
					//r = rgbThemes.at<Vec3f>(0, i)[2];

					//src_intensity = (20.0 * r + 40.0 * g + b) / 61.0;
					//dst_intensity = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);

					//weightMap.at<Vec3f>(j, k)[0] = (float)(dst_intensity * b / src_intensity);
					//weightMap.at<Vec3f>(j, k)[1] = (float)(dst_intensity * g / src_intensity);
					//weightMap.at<Vec3f>(j ,k)[2] = (float)(dst_intensity * r / src_intensity);
				}
			}

			Mat colorWeight = ColorMapJet(grayWeight);
			//colorWeight.convertTo(colorWeight, CV_32FC3, 255.0f);
			imwrite(buf, colorWeight);

			//weightMap.convertTo(weightMap, CV_32FC3, 255.0f);
			//imwrite(buf, weightMap);
		}
	}

	void visWeightMapWithPaletteColor(char *path)
	{
		char buf[256];
		double eps = 1e-8;

		Mat weightMap(height, width, CV_32FC3);
		Mat labtemp(1, 1, CV_32FC3);
		Mat bgrtemp(1, 1, CV_32FC3);

		for (int i = 0; i < k_num; i++)
		{
			sprintf_s(buf, 256, "%s%d.png", path, i+1);

			int sIndex = i * width * height;
			double max_weight = *std::max_element(weight + sIndex, weight + sIndex + width * height);
			double min_weight = *std::min_element(weight + sIndex, weight + sIndex + width * height);

			double r = 0, g = 0, b = 0, src_intensity = 0, dst_intensity = 0;

			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int dIndex = sIndex + j * width + k;
					
					// here we should use the lab color space to visualize the coefficient maps
					float intensity_factor = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);
					labtemp.at<Vec3f>(0, 0)[0] = intensity_factor * labdst.at<Vec3f>(0, i)[0];
					labtemp.at<Vec3f>(0, 0)[1] = labdst.at<Vec3f>(0, i)[1];
					labtemp.at<Vec3f>(0, 0)[2] = labdst.at<Vec3f>(0, i)[2];
					
					cvtColor(labtemp, bgrtemp, CV_Lab2BGR);

					weightMap.at<Vec3f>(j, k)[0] = bgrtemp.at<Vec3f>(0, 0)[0];
					weightMap.at<Vec3f>(j, k)[1] = bgrtemp.at<Vec3f>(0, 0)[1];
					weightMap.at<Vec3f>(j ,k)[2] = bgrtemp.at<Vec3f>(0, 0)[2];

					//b = rgbThemes.at<Vec3f>(0, i)[0];
					//g = rgbThemes.at<Vec3f>(0, i)[1];
					//r = rgbThemes.at<Vec3f>(0, i)[2];

					//src_intensity = (20.0 * r + 40.0 * g + b) / 61.0;
					//dst_intensity = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);

					//if (i == k_num - 1 && dst_intensity >= 0.1)
					//	dst_intensity *= 1.5;

					//weightMap.at<Vec3f>(j, k)[0] = (float)(dst_intensity * b / src_intensity);
					//weightMap.at<Vec3f>(j, k)[1] = (float)(dst_intensity * g / src_intensity);
					//weightMap.at<Vec3f>(j ,k)[2] = (float)(dst_intensity * r / src_intensity);
				}
			}

			weightMap.convertTo(weightMap, CV_32FC3, 255.0f);
			imwrite(buf, weightMap);
		}
	}


	void visualizeCoeffWithRGBPaletteColor(char *path)
	{
		char buf[256];
		double eps = 1e-8;

		Mat weightMap(height, width, CV_32FC3);
		//Mat labtemp(1, 1, CV_32FC3);
		//Mat bgrtemp(1, 1, CV_32FC3);

		int negative_num = 0;

		for (int i = 0; i < k_num; i++)
		{
			sprintf_s(buf, 256, "%s%d.png", path, i+1);

			int sIndex = i * width * height;
			double max_weight = *std::max_element(weight + sIndex, weight + sIndex + width * height);
			double min_weight = *std::min_element(weight + sIndex, weight + sIndex + width * height);

			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int dIndex = sIndex + j * width + k;

					if (weight[dIndex] < 0)
						negative_num++;

					// here we should use the lab color space to visualize the coefficient maps
					float intensity_factor = (weight[dIndex] - min_weight) / (max_weight - min_weight + eps);
					//intensity_factor = intensity_factor < 0 ? 0 : intensity_factor;

					float r = rgbThemes.at<Vec3f>(0, i)[2] * intensity_factor;
					float g = rgbThemes.at<Vec3f>(0, i)[1] * intensity_factor;
					float b = rgbThemes.at<Vec3f>(0, i)[0] * intensity_factor;

					weightMap.at<Vec3f>(j, k)[0] = b;
					weightMap.at<Vec3f>(j, k)[1] = g;
					weightMap.at<Vec3f>(j ,k)[2] = r;

					if (weight[dIndex] < 0)
					{
						weightMap.at<Vec3f>(j, k)[0] = 0.0;
						weightMap.at<Vec3f>(j, k)[1] = 0.0;
						weightMap.at<Vec3f>(j ,k)[2] = 0.0;
					}
				}
			}

			weightMap.convertTo(weightMap, CV_32FC3, 255.0f);
			imwrite(buf, weightMap);
		}

		cout<<"Negative coefficient number: "<<negative_num<<endl;
	}

	void visualizeLayers(char *path)
	{
		char buf[256];
		double eps = 1e-8;

		Mat lab_weightMap(height, width, CV_32FC3);
		Mat bgr_weightMap(height, width, CV_32FC3);

		for (int i = 0; i < k_num; i++)
		{
			sprintf_s(buf, 256, "%s%d.png", path, i+1);

			int sIndex = i * width * height;

			for (int j = 0; j < height; j++)
			{
				for (int k = 0; k < width; k++)
				{
					int dIndex = sIndex + j * width + k;

					// here we should use the lab color space to visualize the coefficient map

					lab_weightMap.at<Vec3f>(j, k)[0] = labdst.at<Vec3f>(0, i)[0];
					lab_weightMap.at<Vec3f>(j, k)[1] = labdst.at<Vec3f>(0, i)[1] * weight[dIndex];
					lab_weightMap.at<Vec3f>(j, k)[2] = labdst.at<Vec3f>(0, i)[2] * weight[dIndex];
					
				}
			}

			cvtColor(lab_weightMap, bgr_weightMap, CV_Lab2BGR);
			bgr_weightMap.convertTo(bgr_weightMap, CV_32FC3, 255.0f);
			imwrite(buf, bgr_weightMap);
		}
	}

private:
	Mat labImg;  
	Mat rgbImg;
	Mat srcImg;
	int k_num;
	int width;
	int height; 

	//CIELAB *labThemes;
	//RGB *rgbThemes;

	double *weight; 
	double *dissimilar_weight;
	Mat rgbThemes;
	Mat labdst;

	Mat channel[3];
};

#endif  // kColorTheme.h