#include "LLE.h"

bool cmp(Manifold a, Manifold b)
{
	return a.indx < b.indx;
}

void LLE::initialization()
{
	dataPts = annAllocPts(pixelNum, dataDim);
	queryPt = annAllocPt(dataDim);

	nnIdx = new ANNidx [kNum+1];
	dists = new ANNdist [kNum+1];

	nnLLEInfo = new Manifold* [pixelNum];
	for (int i = 0; i < pixelNum; i++)
		nnLLEInfo[i] = new Manifold [kNum];

	// loading data points
	for (int i = 0; i < height; i++)
	{
		for (int j = 0; j < width; j++)
		{
			int index = i * width + j;
			Vec3f rgbVal = dstImg.at<Vec3f>(i, j);

			// put the data into dataPts
			//dataPts[index][0] = rgbVal[0];
			//dataPts[index][1] = rgbVal[1];
			//dataPts[index][2] = rgbVal[2];
			//dataPts[index][3] = (float)j / (sigma_p * (width-1));
			//dataPts[index][4] = (float)i / (sigma_p * (height-1));

			dataPts[index][0] = rgbVal[1];
			dataPts[index][1] = rgbVal[2];
			dataPts[index][2] = (float)j / (sigma_p * (width-1));
			dataPts[index][3] = (float)i / (sigma_p * (height-1));
		}
	}

	// construct the kd tree
	kdTree = new ANNkd_tree(dataPts, pixelNum, dataDim);
}

void LLE::KNNSearch()
{
	for (int i = 0; i < height; i++)
	{
		for (int j = 0; j < width; j++)
		{
			int index = i * width + j;
			queryPt = dataPts[index];
			kdTree->annkSearch(queryPt, kNum+1, nnIdx, dists);

			// store the k nearest neighbor 
			for (int h = 0; h < kNum; h++)
				nnLLEInfo[index][h].indx = nnIdx[h+1];
		}
	}

	delete [] nnIdx;
	delete [] dists;
	delete kdTree;
	annClose();
}

void LLE::calcLLEWeights()
{
	double tol = 0.001;
	Mat matData(dataDim, kNum, CV_32FC1);
	Mat tmp(kNum, 1, CV_32FC1, Scalar(1.0));
	Mat Id = Mat::eye(kNum, kNum, CV_32FC1);
	Mat mdx(dataDim, 1, CV_32FC1);
	Mat ex_mdx(dataDim, kNum,CV_32FC1);

	for (int i = 0; i < height; i++)
	{
		for (int j = 0; j < width; j++)
		{
			int index = i * width + j;
			Vec3f rgbVal = dstImg.at<Vec3f>(i, j);

			//for (int dim = 0; dim < 3; dim++)
			//	mdx.at<float>(dim, 0) = rgbVal[dim];
			//mdx.at<float>(3, 0) = (float)j / (sigma_p * (width-1));
			//mdx.at<float>(4, 0) = (float)i / (sigma_p * (height-1));

			for (int dim = 0; dim < 2; dim++)
				mdx.at<float>(dim, 0) = rgbVal[dim+1];
			mdx.at<float>(2, 0) = (float)j / (sigma_p * (width-1));
			mdx.at<float>(3, 0) = (float)i / (sigma_p * (height-1));

			// access the neighbor info of pixel located at (i, j)
			for (int h = 0; h < kNum; h++)
			{
				int idy = nnLLEInfo[index][h].indx / width;
				int idx = nnLLEInfo[index][h].indx % width;

				Vec3f neighborRGBVal = dstImg.at<Vec3f>(idy, idx);

				//for (int ldim = 0; ldim < 3; ldim++)
				//	matData.at<float>(ldim, h) = neighborRGBVal[ldim];
				//matData.at<float>(3, h) = (float)idx / (sigma_p * (width-1));
				//matData.at<float>(4, h) = (float)idy / (sigma_p * (height-1));\

				for (int ldim = 0; ldim < 2; ldim++)
					matData.at<float>(ldim, h) = neighborRGBVal[ldim+1];
				matData.at<float>(2, h) = (float)idx / (sigma_p * (width-1));
				matData.at<float>(3, h) = (float)idy / (sigma_p * (height-1));
			}

			for (int vline = 0; vline < kNum; vline++)
				mdx.col(0).copyTo(ex_mdx.col(vline));

			Mat z = matData - ex_mdx;
			Mat cov = z.t() * z;
			cov = cov + Id * tol * trace(cov)[0];

			Mat wg = cov.inv(DECOMP_LU) * tmp;
			//wg = wg / sum(wg)[0];  
			float nor_sum = 0;
			for (int h = 0; h < wg.rows; h++)
			{
				if (wg.at<float>(h, 0) < 0)
					wg.at<float>(h, 0) = 0;
				else
					nor_sum += wg.at<float>(h, 0);
			}

			for (int h = 0; h < kNum; h++)
				nnLLEInfo[index][h].weight = wg.at<float>(h, 0) / nor_sum;
		}
	}
}

Manifold** LLE::getManifoldPointer()
{
	return nnLLEInfo;
}

void LLE::reArranging()
{
	for (int i = 0; i < height; i++)
	{
		for (int j = 0; j < width; j++)
		{
			int localIndex = i * width + j;
			vector<Manifold> localManifold(nnLLEInfo[localIndex], 
				nnLLEInfo[localIndex] + kNum);
			sort(localManifold.begin(), localManifold.end(), cmp);

			// store sorted manifold 
			for (int h = 0; h < kNum; h++)
				nnLLEInfo[localIndex][h] = localManifold[h];
		}
	}
}
