#ifndef LLE_H
#define LLE_H

#include <cv_header.h>
#include <vector>
#include <ANN/ANN.h>

#pragma comment(lib, "ANN.lib")

#define sigma_p 1.0

typedef struct{
	int indx;
	float weight;
}Manifold;

class LLE{
public:
	LLE(Mat img, int k)
	{
		//dataDim = 5;  // (l, a, b, x, y)
		dataDim = 4;   // (a, b, x, y)

		kNum = k;
		width  = img.cols;
		height = img.rows;

		Mat uchar_labImg;
		cvtColor(img, uchar_labImg, CV_BGR2Lab);
		uchar_labImg.convertTo(dstImg, CV_32FC3, 1/255.0f);
		//img.convertTo(dstImg, CV_32FC3, 1/255.0f);

		pixelNum = width * height;
	}
	~LLE()
	{
		for (int i = 0; i < pixelNum; i++)
			delete [] nnLLEInfo[i];
		delete [] nnLLEInfo;
	}

public:
	void initialization();
	void KNNSearch();
	void calcLLEWeights();
	void reArranging();
	Manifold** getManifoldPointer();

private:
	int kNum;
	int width;
	int height;
	int pixelNum;  
	int dataDim;
	Mat dstImg;

	Manifold **nnLLEInfo;

	ANNpointArray dataPts;    // data points
	ANNpoint queryPt;          // query point
	ANNidxArray  nnIdx;       // nearest neighbor indices
	ANNdistArray  dists;        // near neighbor distances
	ANNkd_tree*   kdTree;     // search structure
};

#endif