#include "calcAffinity.h"

void calcAffinityArray( Mat src, vector<vector<AffinityArray> > &weights )
{
	Mat labImg;
	cvtColor(src, labImg, CV_BGR2Lab);

	int width = src.cols;
	int height = src.rows;
	int pageCount = width * height;

	float eps = 1e-8;
	for (int i = 0; i < height; i++)
	{
		for (int j = 0; j < width; j++)
		{
			int global_index = i * width + j;
			int top_left_x = j > RADIUS ? (j - RADIUS) : 0;
			int top_left_y = i > RADIUS ? (i - RADIUS) : 0;
			int bottom_right_x = (j + RADIUS) > width ? (width - 1) : (j + RADIUS);
			int bottom_right_y = (i + RADIUS) > height ? (height - 1) : (i + RADIUS);

			Vec3b labVal = labImg.at<Vec3b>(i, j);

			float norm_weight = eps;

			for (int ii = top_left_y; ii <= bottom_right_y; ii++)
			{
				for (int jj = top_left_x; jj <= bottom_right_x; jj++)
				{
					if (ii == i && jj == j)
						continue;

					int local_index = ii * width + jj;
					Vec3b local_labVal = labImg.at<Vec3b>(ii, jj);
					float color_sqDist = (float)(labVal[2] - local_labVal[2]) * (labVal[2] - local_labVal[2]) + 
						(float)(labVal[1] - local_labVal[1]) * (labVal[1] - local_labVal[1]) + 
						(float)(labVal[0] - local_labVal[0]) * (labVal[0] - local_labVal[0]);
					float spatial_sqDist = (float) (ii - i) * (ii - i) / ((height - 1) * (height - 1)) + 
						(float) (jj - j) * (jj - j) / ((width - 1) * (width - 1));

					float local_weight = exp(-color_sqDist / sigma_a) * exp(-spatial_sqDist / sigma_s);
					AffinityArray Aff;
					Aff.weight = local_weight;
					Aff.ind = local_index;
					weights[global_index].push_back(Aff);

					norm_weight += local_weight;
				}
			}

			// normalize the weight
			for (int h = 0; h < weights[global_index].size(); h++)
			{
				weights[global_index][h].weight /= norm_weight;
			}
		}
	}

}
